# init files are still required in modern python!
# https://peps.python.org/pep-0420/ introduced implicit namespace packages
# but for building and many toolings, you still need to have __init__ files (at least in the root of the package).
# e.g. if you remove this init file and try to build with pip install .
# you won't be able to import the dummy module.



"""merge 2 coco datasets into one"""

import json
import pathlib
import shutil

import tqdm
from airo_dataset_tools.data_parsers.coco import CocoInstancesDataset


def merge_coco_annotations(dataset1: CocoInstancesDataset, dataset2: CocoInstancesDataset) -> CocoInstancesDataset:
    """merge 2 coco annotations schemas. Categories and Annotations are assumed to be unique. Images can be present in both datasets.

    Images will be checked for duplicates based on their file name.

    Annotation IDs will be changed to avoid conflicts and their image IDs will be updated if needed."""
    categories_1 = dataset1.categories
    categories_2 = dataset2.categories

    merged_categories = list(categories_1)
    for category in categories_2:
        if category.id not in [category.id for category in merged_categories]:
            merged_categories.append(category)
        else:
            same_id_category = merged_categories[[category.id for category in merged_categories].index(category.id)]
            if category != same_id_category:
                raise ValueError(
                    "Categories with the same ID are not equal: " + str(category) + " " + str(same_id_category)
                )

    merged_images = dataset1.images
    max_dataset_image_id = max([image.id for image in merged_images])
    merged_image_paths = [image.file_name for image in merged_images]

    dataset_2_image_id_mapping = {}
    for image in dataset2.images:
        if image.file_name not in merged_image_paths:
            # create a new ID to avoid collisions
            old_id = image.id
            image.id = max_dataset_image_id + 1
            max_dataset_image_id += 1
            dataset_2_image_id_mapping[old_id] = image.id
            merged_images.append(image)
            merged_image_paths.append(image.file_name)
        else:
            # dataset already contained this image, so we need to remap the annotations
            dataset_2_image_id_mapping[image.id] = merged_images[merged_image_paths.index(image.file_name)].id

    merged_annotations = list(dataset1.annotations)
    max_dataset_annotation_id = max([annotation.id for annotation in merged_annotations])
    for annotation in dataset2.annotations:
        if annotation.image_id in dataset_2_image_id_mapping.keys():
            annotation.image_id = dataset_2_image_id_mapping[annotation.image_id]
        annotation.id = max_dataset_annotation_id + 1
        max_dataset_annotation_id += 1
        merged_annotations.append(annotation)

    merged_dataset = CocoInstancesDataset(
        categories=merged_categories, images=merged_images, annotations=merged_annotations
    )
    return merged_dataset


def merge_coco_image_folders(dataset1_base_dir: str, dataset2_base_dir: str, target_dir: str) -> None:
    """merge 2 image folders into one. Images will be copied to the target folder.

    all base_dirs have following setup:

    base_dir
    --- images
    ------ image1.<>
    ------ image2.<>
    <>.json // annotations with path relative to base_dir
    """

    dataset1_base_dir_path = pathlib.Path(dataset1_base_dir)
    dataset2_base_dir_path = pathlib.Path(dataset2_base_dir)
    target_dir_path = pathlib.Path(target_dir)

    dataset1_image_paths = [image_path for image_path in dataset1_base_dir_path.iterdir()]
    dataset2_image_paths = [image_path for image_path in dataset2_base_dir_path.iterdir()]

    target_image_dir = target_dir_path / "images"
    target_image_dir.mkdir(parents=True, exist_ok=True)

    for image_path in tqdm.tqdm(
        dataset1_image_paths, desc=f"copying images from {dataset1_base_dir_path.name} to {target_dir_path.name}"
    ):
        shutil.copy(image_path, target_image_dir / image_path.name)

    for image_path in tqdm.tqdm(
        dataset2_image_paths, desc=f"copying images from {dataset2_base_dir_path.name} to {target_dir_path.name}"
    ):
        if not (target_image_dir / image_path.name).exists():
            shutil.copy(image_path, target_image_dir / image_path.name)


def merge_coco_datasets(json_path_1: str, json_path_2: str, target_json_path: str) -> None:
    """merge 2 coco datasets into one. Categories and Annotations are assumed to be unique. Images can be present in both datasets.

    Images will be checked for duplicates based on their file name.

    Annotation IDs will be changed to avoid conflicts and their image IDs will be updated if needed."""

    image_path_1 = pathlib.Path(json_path_1).parent / "images"
    image_path_2 = pathlib.Path(json_path_2).parent / "images"

    merge_coco_image_folders(str(image_path_1), str(image_path_2), str(pathlib.Path(target_json_path).parent))

    dataset1 = None
    with open(json_path_1, "r") as f:
        dataset1 = CocoInstancesDataset(**json.load(f))

    dataset2 = None
    with open(json_path_2, "r") as f:
        dataset2 = CocoInstancesDataset(**json.load(f))

    merged_dataset = merge_coco_annotations(dataset1, dataset2)

    with open(target_json_path, "w") as f:
        json.dump(merged_dataset.model_dump(exclude_none=True), f)


if __name__ == "__main__":
    json1 = (
        "/home/tlips/Documents/airo-mono/airo-dataset-tools/test/test_data/lego-battery-resized/annotations_train.json"
    )
    json2 = (
        "/home/tlips/Documents/airo-mono/airo-dataset-tools/test/test_data/lego-battery-resized/annotations_val.json"
    )
    target_json = "/home/tlips/Documents/airo-mono/airo-dataset-tools/test/test_data/lego-battery-resize-merged/annotations_merged.json"
    merge_coco_datasets(json1, json2, target_json)



from __future__ import annotations

from typing import Any, List, Optional, Tuple

import cv2
import numpy as np
from airo_dataset_tools.data_parsers.coco import Polygon, RLEDict, Segmentation
from pycocotools import mask


def merge_multi_segment(segments: List[List[Any]]) -> List[np.ndarray]:
    """

    code taken from: https://github.com/ultralytics/JSON2YOLO/blob/main/general_json2yolo.py#L330
    (AGPL licensed.)

    Merge multi segments to one list. Find the coordinates with min distance between each segment, then connect these
    coordinates with one thin line to merge all segments into one.

    This is useful to convert masks to yolo dataset, as yolo only supports one polygon for each object.

    Args:
        segments(List(List)): original segmentations in coco's json file.
            like [segmentation1, segmentation2,...],
            each segmentation is a list of coordinates.

    Returns:
        List(List): merged segments.
    """

    def min_index(arr1: np.ndarray, arr2: np.ndarray) -> Tuple[Any, ...]:
        """
        Find a pair of indexes with the shortest distance.

        Args:
            arr1: (N, 2).
            arr2: (M, 2).
        Return:
            a pair of indexes(tuple).
        """
        dis = ((arr1[:, None, :] - arr2[None, :, :]) ** 2).sum(-1)
        return np.unravel_index(np.argmin(dis, axis=None), dis.shape)

    s = []
    segments_arrays = [np.array(i).reshape(-1, 2) for i in segments]
    idx_list: List[List[int]] = [[] for _ in range(len(segments_arrays))]

    # record the indexes with min distance between each segment
    for i in range(1, len(segments_arrays)):
        idx1, idx2 = min_index(segments_arrays[i - 1], segments_arrays[i])
        idx_list[i - 1].append(idx1)
        idx_list[i].append(idx2)

    # use two round to connect all the segments
    for k in range(2):
        # forward connection
        if k == 0:
            for i, idx in enumerate(idx_list):
                # middle segments have two indexes
                # reverse the index of middle segments
                if len(idx) == 2 and idx[0] > idx[1]:
                    idx = idx[::-1]
                    segments_arrays[i] = segments_arrays[i][::-1, :]

                segments_arrays[i] = np.roll(segments_arrays[i], -idx[0], axis=0)
                segments_arrays[i] = np.concatenate([segments_arrays[i], segments_arrays[i][:1]])
                # deal with the first segment and the last one
                if i in [0, len(idx_list) - 1]:
                    s.append(segments_arrays[i])
                else:
                    idx = [0, idx[1] - idx[0]]
                    s.append(segments_arrays[i][idx[0] : idx[1] + 1])

        else:
            for i in range(len(idx_list) - 1, -1, -1):
                if i not in [0, len(idx_list) - 1]:
                    idx = idx_list[i]
                    nidx = abs(idx[1] - idx[0])
                    s.append(segments_arrays[i][nidx:])

    s = np.concatenate(s, axis=0).reshape(-1).tolist()
    return s


class BinarySegmentationMask:
    """Class that holds a binay segmentation mask and can convert between binary bitmask and/or the different COCO segmentation formats:
    - polygon: [list[list[float]]] containing [x,y] coordinates of the polygon(s)
    - uncompressed RLE: [dict] with keys "counts" and "size" where count contains the actual run length encoding[x1,l1,x2,l2,...]
    - compressed RLE: [dict] with keys "counts" and "size" where count contains a coco-encoded string that represents the run length encoding
    """

    def __init__(self, bitmap: np.ndarray):
        assert np.array_equal((bitmap == 1.0) * 1.0, bitmap), "bitmap must be 1 or 0 numpy array"
        bitmap = bitmap.astype(np.uint8)
        self.bitmap = bitmap

    @classmethod
    def from_coco_segmentation_mask(
        cls, segmentation: Segmentation, width: int, height: int
    ) -> BinarySegmentationMask:
        """Convert a coco segmentation mask to a bitmap. based on coco"""

        # convert to encoded RLE if required
        if isinstance(segmentation, list):
            # polygon [list[list[float]]]
            rles = mask.frPyObjects(segmentation, height, width)
            rle = mask.merge(rles)
        elif isinstance(segmentation, dict):
            if isinstance(segmentation["counts"], list):
                # uncompressed RLE
                rle = mask.frPyObjects(segmentation, height, width)
            else:
                # encoded RLE
                rle = segmentation

        # decode encoded RLE to bitmap
        else:
            raise ValueError("segmentation must be a valid coco segmentation mask")
        bitmap = mask.decode(rle)
        return BinarySegmentationMask(bitmap)

    @property
    def area(self) -> float:
        return float(mask.area(self.as_compressed_rle))

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        """returns x,y,w,h of the enclosing bbox"""

        bbox: np.ndarray = mask.toBbox(self.as_compressed_rle)
        return (bbox[0], bbox[1], bbox[2], bbox[3])

    @property
    def as_polygon(self) -> Optional[List[Polygon]]:
        # from https://github.com/cocodataset/cocoapi/issues/476#issuecomment-871804850
        contours, _ = cv2.findContours(self.bitmap, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        segmentation = []
        valid_poly = 0
        for contour in contours:
            # Valid polygons have >= 6 coordinates (3 points)
            if contour.size >= 6:
                segmentation.append(contour.astype(float).flatten().tolist())
                valid_poly += 1
        if valid_poly == 0:
            print("No valid polygons found in segmentation mask")
            return None
        return segmentation

    @property
    def as_single_polygon(self) -> Optional[Polygon | List[np.ndarray]]:
        """Convert a bitmap to a single polygon. If the bitmap contains multiple segments, they will be merged into one polygon."""
        poly = self.as_polygon
        if poly is None:
            return None

        if len(poly) == 1:
            return poly[0]

        poly_merged = merge_multi_segment(poly)
        return poly_merged

    @property
    def as_uncompressed_rle(self) -> RLEDict:
        raise NotImplementedError

    @property
    def as_compressed_rle(self) -> RLEDict:
        """Convert a bitmap to a compressed coco RLEDict"""
        b = np.asfortranarray(self.bitmap.astype(np.uint8))
        encoded_rle: RLEDict = mask.encode(b)
        assert isinstance(encoded_rle["counts"], bytes)
        encoded_rle["counts"] = encoded_rle["counts"].decode("utf-8")
        return encoded_rle


if __name__ == "__main__":
    m = np.zeros((10, 10)).astype(np.uint8)
    m[1:3] = 1
    m[6:8, 1:5] = 1
    print(m)
    print(BinarySegmentationMask(m).as_polygon)
    poly = BinarySegmentationMask(m).as_single_polygon
    # poly = merge_multi_segment(poly)
    print(poly)
    mask2 = BinarySegmentationMask.from_coco_segmentation_mask(poly, 10, 10)  # type: ignore # TODO fix this
    print(mask2.bitmap)



""" Split a COCO dataset into subsets"""
import json
import pathlib
import random
from typing import List, Optional

from airo_dataset_tools.data_parsers.coco import CocoImage, CocoInstanceAnnotation, CocoInstancesDataset
from pydantic import ValidationError


def split_coco_dataset(
    coco_dataset: CocoInstancesDataset,
    split_ratios: List[float],
    shuffle_before_splitting: bool = True,
    shuffle_seed: int = 42,
) -> List[CocoInstancesDataset]:
    """Split a COCO dataset into subsets by splitting the images according to the specified relative ratios.
    All annotations for an image will be placed in the same subset as the image.

    Note that this does not guarantee the ratio of the annotations OR an equal class balance in each subset.

    Ratios must sum to 1.0.
    """
    ratio_sum = sum(split_ratios)
    if abs(ratio_sum - 1.0) > 2e-2:
        raise ValueError(f"Ratios must sum to 1.0. Ratios sum to {ratio_sum}.")

    # split the images into 2 subsets (random or ordered)
    images = coco_dataset.images

    if shuffle_before_splitting:
        # make shuffling reproducible
        # use local random instance to not influence
        # the global random state
        rng = random.Random(shuffle_seed)
        rng.shuffle(images)

    image_splits: List[List[CocoImage]] = []
    split_sizes = [round(ratio * len(images)) for ratio in split_ratios]
    split_sizes[-1] = len(images) - sum(split_sizes[:-1])  # make sure the total number of images is correct
    print(f"Split sizes: {split_sizes}")

    for size in split_sizes:
        image_splits.append(images[:size])
        images = images[size:]

    image_id_to_split_id = {}
    for split_id, image_split in enumerate(image_splits):
        for image in image_split:
            image_id_to_split_id[image.id] = split_id

    # gather the annotations for each subset
    annotation_splits: List[List[CocoInstanceAnnotation]] = [[] for _ in range(len(split_ratios))]
    for annotation in coco_dataset.annotations:
        image_id = annotation.image_id
        split_id = image_id_to_split_id[image_id]
        # keep original image_ids and annotation_ids so that you could still reference the original dataset
        annotation_splits[split_id].append(annotation)

    # check if none of the annotation splits are empty:
    for split_id, annotation_split in enumerate(annotation_splits):
        if len(annotation_split) == 0:
            raise ValueError(
                f"Split {split_id} is empty, which is not allowed. Please use a larger dataset or a smaller split ratio."
            )
    # create a new COCO dataset for each subset
    coco_dataset_splits: List[CocoInstancesDataset] = []
    for annotation_split, image_split in zip(annotation_splits, image_splits):
        coco_dataset_split = CocoInstancesDataset(
            categories=coco_dataset.categories, images=image_split, annotations=annotation_split
        )
        coco_dataset_splits.append(coco_dataset_split)

    return coco_dataset_splits


def split_and_save_coco_dataset(
    coco_json_path: str, split_ratios: List[float], shuffle_before_splitting: bool = True
) -> None:
    """Split a COCO dataset into subsets according to the specified relative ratios and save them to disk.
    Images are split with their corresponding annotations. No guarantees on class balance or annotation ratios.

    Ratios must sum to 1.0.

    If two ratios are specified, the dataset will be split into two subsets. these will be called train/val by default.
    If three ratios are specified, the dataset will be split into three subsets. these will be called train/val/test by default.
    """
    split_names = ["train", "val", "test"]
    if len(split_ratios) > len(split_names):
        raise ValueError(f"Only {len(split_names)} splits are supported. {len(split_ratios)} splits were specified.")

    coco_dataset: Optional[CocoInstancesDataset] = None
    with open(coco_json_path, "r") as f:
        try:
            coco_dataset = CocoInstancesDataset(**json.load(f))
        except ValidationError as e:
            print(e)
            raise ValueError("Could not load CocoInstancesDataset")

    coco_dataset_splits = split_coco_dataset(coco_dataset, split_ratios, shuffle_before_splitting)

    for split_id, coco_dataset_split in enumerate(coco_dataset_splits):

        file_name = coco_json_path.replace(".json", f"_{split_names[split_id]}.json")
        with open(file_name, "w") as f:
            json.dump(coco_dataset_split.model_dump(exclude_none=True), f)


if __name__ == "__main__":
    json_path = pathlib.Path(__file__).parents[2] / "test" / "test_data" / "person_keypoints_val2017_small.json"
    print(json_path)
    split_and_save_coco_dataset(str(json_path), [0.5, 0.5])



"""Tests for loading datasets in COCO format.

Part of the tests use annotations from the official COCO dataset.
For this we use a small subset of the 2017 Val annotations files, the full versions can be found here:
https://cocodataset.org/#download

We do not use the captions, so we only consider these files:
- instances_val2017.json
- person_keypoints_val2017.json
"""

import json
import os

import pytest
from airo_dataset_tools.data_parsers.coco import (
    CocoCategory,
    CocoInstancesDataset,
    CocoKeypointCategory,
    CocoKeypointsDataset,
)


def test_coco_load_instances():
    """Test whether we can correctly load the partial instances_val2017 dataset."""
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/instances_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_instances = CocoInstancesDataset(**data)
        # Check a few known lengths to ensure that all elements were loaded
        assert len(coco_instances.images) == 3
        assert len(coco_instances.categories) == 80
        assert len(coco_instances.annotations) == 34

        assert isinstance(coco_instances.categories[0], CocoCategory)


def test_coco_load_keypoints():
    """Test whether we can correctly load the partial person_keypoints_val2017 dataset."""
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_keypoints = CocoKeypointsDataset(**data)
        assert len(coco_keypoints.images) == 2
        assert len(coco_keypoints.categories) == 1
        assert len(coco_keypoints.annotations) == 5

        assert isinstance(coco_keypoints.categories[0], CocoKeypointCategory)


def test_coco_load_keypoints_as_instances_keeps_additional_fields():
    # to parse data that actually belongs to a subclass, useful for some coco tools where it does not matter what category the dataset is..
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_keypoints = CocoInstancesDataset(**data)
        assert len(coco_keypoints.annotations[0].keypoints) > 0


def test_coco_load_instances_incorrectly():
    """Test whether an exception is raised when we try to load the regular instances as a keypoints dataset."""
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/instances_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        with pytest.raises(Exception):
            CocoKeypointsDataset(**data)


def test_coco_load_instances_no_segmasks():
    """Test whether we can correctly load the partial instances_val2017 dataset without segmentations."""
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small_no_segmentations.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_instances = CocoKeypointsDataset(**data)
        # Check a few known lengths to ensure that all elements were loaded
        assert len(coco_instances.images) == 2


def test_coco_load_keypoints_no_bboxes():
    """Test whether we can correctly load the partial person_keypoints_val2017 dataset without bounding boxes."""
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small_no_bbox.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_keypoints = CocoKeypointsDataset(**data)
        assert len(coco_keypoints.images) == 2
        assert len(coco_keypoints.categories) == 1
        assert len(coco_keypoints.annotations) == 2

        assert isinstance(coco_keypoints.categories[0], CocoKeypointCategory)



import os
from typing import List, Optional


def view_coco_dataset(
    labels_json_path: str, dataset_dir: Optional[str] = None, label_types: Optional[List[str]] = None
) -> None:
    """visualize a coco dataset in fiftyone"""
    # lazy import because fiftyone is slow to import and makes the CLI slow.
    import fiftyone as fo

    if dataset_dir is None:
        dataset_dir = os.path.dirname(labels_json_path)
    if label_types is None or not label_types:
        label_types = ["detections", "segmentations", "keypoints"]
    else:
        assert all([label_type in ["detections", "segmentations", "keypoints"] for label_type in label_types])

    labels_json_path = os.path.realpath(labels_json_path)

    dataset = fo.Dataset.from_dir(
        dataset_type=fo.types.COCODetectionDataset,
        label_types=label_types,
        data_path=dataset_dir,
        labels_path=labels_json_path,
    )

    session = fo.launch_app(dataset)
    session.wait()



import xmltodict


def get_dict_from_xml(xml_path: str) -> dict:
    with open(xml_path, "r") as file:
        # prefixes @/_ lead to issues with pydantic parsing! so simply use no prefix
        xml_dict = xmltodict.parse(file.read(), attr_prefix="")
        return xml_dict


if __name__ == "__main__":
    """convert XML to JSON"""
    import json
    import sys

    assert len(sys.argv) == 2
    xml_path = sys.argv[1]
    json_path = xml_path[:-3] + "json"

    print(f"converting {xml_path} to {json_path}")

    xml_dict = get_dict_from_xml(xml_path)
    with open(json_path, "w") as outfile:
        json.dump(xml_dict, outfile)



"""Tests for creating datasets in COCO format."""

from airo_dataset_tools.data_parsers.coco import (
    CocoCategory,
    CocoImage,
    CocoInfo,
    CocoInstanceAnnotation,
    CocoInstancesDataset,
    CocoKeypointAnnotation,
    CocoKeypointCategory,
    CocoKeypointsDataset,
    CocoLicense,
)


def coco_info_example() -> CocoInfo:
    return CocoInfo(
        description="Description of the dataset.",
        url="https://github.com/airo-ugent/airo-mono",
        version="0.1",
        year=2023,
        contributor="Victor-Louis De Gusseme",
        date_created="2023/03/06",
    )


def coco_image_example() -> CocoImage:
    return CocoImage(
        id=1,
        width=640,
        height=480,
        file_name="test_image.jpg",
        license=1,
        flicker_url="https://flicker.com/test_image.jpg",
        coco_url="https://coco.com/test_image.jpg",
        date_captured="2023/03/06",
    )


def coco_image_without_optional_fields_example() -> CocoImage:
    return CocoImage(
        id=1,
        width=640,
        height=480,
        file_name="test_image.jpg",
    )


def coco_category_example() -> CocoCategory:
    return CocoCategory(
        supercategory="cloth",
        id=99,
        name="towel",
    )


def coco_keypoint_category_example() -> CocoKeypointCategory:
    return CocoKeypointCategory(
        supercategory="cloth",
        id=99,
        name="towel",
        keypoints=["corner_1", "corner_2", "corner_3"],
    )


def coco_instance_annotation_example() -> CocoInstanceAnnotation:
    return CocoInstanceAnnotation(
        id=1,
        image_id=1,
        category_id=99,
        segmentation=[[0, 0, 50, 75, 100, 0]],
        area=375.0,
        bbox=(0, 0, 100, 75),
        iscrowd=0,
    )


def coco_instance_annotation_with_rle_example() -> CocoInstanceAnnotation:
    # TODO: make segmentation, area and bbox consistent
    return CocoInstanceAnnotation(
        id=2,
        image_id=1,
        category_id=99,
        segmentation={"counts": [0, 10, 1, 20], "size": [100, 75]},
        area=375.0,
        bbox=(0, 0, 100, 75),
        iscrowd=0,
    )


def coco_keypoint_annotation_example() -> CocoKeypointAnnotation:
    return CocoKeypointAnnotation(
        id=3,
        image_id=1,
        category_id=99,
        segmentation=[[0, 0, 50, 75, 100, 0]],
        area=375.0,
        bbox=(0, 0, 100, 75),
        iscrowd=0,
        keypoints=[0, 0, 1, 50, 75, 2, 100, 0, 0],
        num_keypoints=2,
    )


def coco_license_example() -> CocoLicense:
    return CocoLicense(
        id=1,
        name="Attribution-NonCommercial-ShareAlike License",
        url="http://creativecommons.org/licenses/by-nc-sa/2.0/",
    )


def test_coco_info():
    assert coco_info_example()


def test_coco_image():
    assert coco_image_example()


def test_coco_image_without_optional_fields():
    assert coco_image_without_optional_fields_example()


def test_coco_category():
    assert coco_category_example()


def test_coco_keypoint_category():
    assert coco_keypoint_category_example()


def test_coco_instance_annotation():
    assert coco_instance_annotation_example()


def test_coco_instance_annotation_with_rle():
    assert coco_instance_annotation_with_rle_example()


def test_coco_keypoint_annotation():
    assert coco_keypoint_annotation_example()


def test_coco_license():
    assert coco_license_example()


def test_coco_instances():
    coco_info = coco_info_example()
    coco_license = coco_license_example()
    coco_image = coco_image_example()
    coco_category = coco_category_example()
    coco_instance_annotation = coco_instance_annotation_example()
    coco_instance_annotation_with_rle = coco_instance_annotation_with_rle_example()

    coco_instances = CocoInstancesDataset(
        info=coco_info,
        licenses=[coco_license],
        images=[coco_image],
        categories=[coco_category],
        annotations=[coco_instance_annotation, coco_instance_annotation_with_rle],
    )
    assert coco_instances


def test_coco_keypoints():
    coco_info = coco_info_example()
    coco_license = coco_license_example()
    coco_image = coco_image_example()
    coco_keypoint_category = coco_keypoint_category_example()
    coco_keypoint_annotation = coco_keypoint_annotation_example()

    coco_keypoints = CocoKeypointsDataset(
        info=coco_info,
        licenses=[coco_license],
        images=[coco_image],
        categories=[coco_keypoint_category],
        annotations=[coco_keypoint_annotation],
    )
    assert coco_keypoints



from airo_dataset_tools.cvat_labeling.convert_cvat_to_coco import cvat_image_to_coco

from .test_cvat_images_load import COCO_CATEGORIES_PATH, CVAT_EXAMPLE_PATH


def test_conversion():
    coco_dict = cvat_image_to_coco(CVAT_EXAMPLE_PATH, COCO_CATEGORIES_PATH, add_bbox=True, add_segmentation=True)
    assert len(coco_dict["images"]) == 4



import json
import os

import numpy as np
from airo_dataset_tools.data_parsers.coco import CocoInstancesDataset
from airo_dataset_tools.segmentation_mask_converter import BinarySegmentationMask


def test_encoded_rle_creation():
    mask = np.zeros((10, 10))
    mask[0, 0] = 1
    mask[1, 1:3] = 1
    mask[2, 2:4] = 1
    segmentation_mask = BinarySegmentationMask(mask)

    compressed_rle = segmentation_mask.as_compressed_rle
    assert isinstance(compressed_rle, dict)
    assert "counts" in compressed_rle.keys()
    assert "size" in compressed_rle.keys()
    assert compressed_rle["size"] == [10, 10]
    assert isinstance(compressed_rle["counts"], str)

    loaded_segmentation_mask = BinarySegmentationMask.from_coco_segmentation_mask(compressed_rle, 10, 10)
    assert np.array_equal(loaded_segmentation_mask.bitmap, mask)


def test_polygon_creation():
    mask = np.zeros((10, 10))
    mask[1, 1:3] = 1
    mask[2, 1:3] = 1
    segmentation_mask = BinarySegmentationMask(mask)

    polygon = segmentation_mask.as_polygon
    assert isinstance(polygon, list)
    assert isinstance(polygon[0], list)
    assert isinstance(polygon[0][0], float)
    assert len(polygon) == 1
    assert len(polygon[0]) == 8
    assert polygon[0] == [1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 1.0]


def test_coco_segmentation_loading():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/instances_val2017_small.json")
    with open(annotations, "r") as file:
        data = json.load(file)
        coco_instances = CocoInstancesDataset(**data)
        for annotation in coco_instances.annotations:
            for image in coco_instances.images:
                if image.id == annotation.image_id:
                    width = image.width
                    height = image.height
            segmentation_mask = BinarySegmentationMask.from_coco_segmentation_mask(
                annotation.segmentation, width, height
            )
            assert isinstance(segmentation_mask.bitmap, np.ndarray)
            assert segmentation_mask.bitmap.shape == (height, width)


if __name__ == "__main__":
    test_encoded_rle_creation()
    test_coco_segmentation_loading()



"""Parser for CVAT Images 1.1 keypoint annotations

see CVAT Documentation for format information: https://opencv.github.io/cvat/docs/manual/advanced/xml_format/

This file was created by labeling the example images using the flow that is described in the repo.
Then this xml was converted to JSON using xmltodict
Then an initial version of the parser was created using https://pydantic-docs.helpmanual.io/datamodel_code_generator/:
`datamodel-codegen --input annotations.json --input-file-type json --output cvat_keypoints_parser.py --class-name CVATKeypointsParser`

And the parser was then further finetuned.

Note that the attributes cannot have _ as prefix for Pydantic.
"""

# generated by datamodel-codegen:
#   filename:  annotations.json
#   timestamp: 2022-08-24T12:15:52+00:00
# then further changed by @tlips

from __future__ import annotations

from typing import Any, List, Optional, Union

from pydantic import BaseModel, field_validator


class Segment(BaseModel):
    id: str
    start: str
    stop: str
    url: str


class Segments(BaseModel):
    segment: Segment


class Owner(BaseModel):
    username: str
    email: str


class LabelItem(BaseModel):
    name: str
    color: str
    type: str
    attributes: Any


class Labels(BaseModel):
    label: Union[List[LabelItem], LabelItem]


class Job(BaseModel):
    id: str
    name: Optional[str] = None
    size: str
    mode: str
    overlap: str
    bugtracker: Any
    created: str
    updated: str
    subset: str
    start_frame: str
    stop_frame: str
    frame_filter: Any
    segments: Segments
    owner: Owner
    assignee: Any
    labels: Labels


class Meta(BaseModel):
    job: Optional[Job] = None
    task: Optional[Job] = None
    dumped: str

    def get_job_or_task(self) -> Job:
        if self.job:
            return self.job
        elif self.task:
            return self.task
        else:
            raise ValueError("No task or job found in meta. This should not happen")


class AnnotationItem(BaseModel):
    label: str
    source: str
    z_order: str
    occluded: str
    group_id: Optional[str] = "1"  # set default group id to 1.


class Point(AnnotationItem):
    points: str

    @field_validator("points")
    @classmethod
    def has_two_coordinates(cls, v: str) -> str:
        assert len(v.split(",")) == 2, "each point must be a single 2D coordinate for the AIRO flow."
        return v


class Polygon(AnnotationItem):
    points: str


class Box(AnnotationItem):
    xtl: str
    ytl: str
    xbr: str
    ybr: str


class ImageItem(BaseModel):
    id: str
    name: str
    width: str
    height: str
    points: Optional[Union[List[Point], Point]] = None
    polygon: Optional[Union[List[Polygon], Polygon]] = None
    box: Optional[Union[List[Box], Box]] = None


class Annotations(BaseModel):
    version: str
    meta: Meta
    image: List[ImageItem]


class CVATImagesParser(BaseModel):
    annotations: Annotations


if __name__ == "__main__":
    import json
    import pathlib

    path = pathlib.Path(__file__).parent.absolute()
    cvat_dict = json.load(open(str(path.parent / "cvat_labeling" / "example" / "annotations.json")))
    cvat_keypoints_parser = CVATImagesParser(**cvat_dict)



"""parser for COCO dataset (both instance & keypoints) JSON files.

The main reference for this parser can be found here:
https://cocodataset.org/#format-data

Dataset loading example:

with open("path/to/annotations.json","r") as file:
    data = json.load(file) # dict
    parsed_data = CocoKeypoints(**data)

When there are no keypoints, you may use CocoInstances instead of CocoKeypoints.


Dataset creation example:

coco_image = CocoImage(
    id=1,
    width=100,
    height=100,
    file_name="path/to/image.jpg",
)
images = [coco_image]

# make more images, catergories and annotations here

coco_keypoints = CocoKeypoints(categories=categories, images=images, annotations=annotations)
with open("annotations.json", "w") as file:
    json.dump(coco_keypoints.model_dump(exclude_none=True), file, indent=4)

"""

from typing import Dict, List, Optional, Sequence, Tuple, Union

from pydantic import BaseModel, field_validator, model_validator

# Used by CocoInfo and CocoImage
Datetime = Union[
    str, int
]  # COCO uses both YYYY-MM-DD and YYYY-MM-DD HH:MM:SS for datetime, CVAT places 0 for missing values
Url = str

# Used by CocoImage and CocoLicenses
LicenseID = int

# Used by CocoCategory and CocoInstanceAnnotation
CategoryID = int

# Used by CocoImage and CocoInstanceAnnotation
ImageID = int

# Used by CocoInstanceAnnotation
# Dict[int,int]  run length encoding(RLE) (of a pixel-mask), with keys "counts" and "size"
# where count contains the actual run length encoding (of a pixel-mask) [x1,l1,x2,l2,...]
# or a coco-encoded string that represents this list.
# see https://www.youtube.com/watch?v=h6s61a_pqfM for a video explanation of RLE
# see https://github.com/cocodataset/cocoapi/issues/386 for a discussion of RLE string encoding
# RLE is Usually only used for is_crowd segmentations.
RLEDict = Dict[str, Union[str, List]]

Polygon = List[float]  # list of vertices [x1, y1, x2, y2, ...]
Segmentation = Union[RLEDict, List[Polygon]]

# Used by the Annotations
Keypoints = List[float]  # list of keypoints [x1, y1, v1, x2, y2, v2, ...]
IsCrowd = int  # 0 or 1


class CocoInfo(BaseModel):
    year: Union[int, str]
    version: str
    description: str
    contributor: str
    url: Url
    date_created: Datetime


class CocoImage(BaseModel):
    id: ImageID
    width: int
    height: int
    file_name: str  # relative path to image
    license: Optional[LicenseID] = None
    flicker_url: Optional[Url] = None
    coco_url: Optional[Url] = None
    date_captured: Optional[Datetime] = None
    __hash__ = object.__hash__  # make hashable for use in set


class CocoCategory(BaseModel, extra="allow"):

    supercategory: str  # should be set to "name" for root category
    id: CategoryID
    name: str


class CocoKeypointCategory(CocoCategory):
    keypoints: List[str]
    skeleton: Optional[List[List[int]]] = None


class CocoInstanceAnnotation(BaseModel, extra="allow"):  # allow extra fields, to parse subclasses

    id: int  # unique id for the annotation
    image_id: ImageID
    category_id: CategoryID

    bbox: Optional[Tuple[float, float, float, float]] = None

    segmentation: Optional[Segmentation] = None
    area: Optional[float] = None
    iscrowd: Optional[IsCrowd] = None

    @field_validator("iscrowd")
    @classmethod
    def iscrowd_must_be_binary(cls, v: IsCrowd) -> IsCrowd:
        if v is None:
            return None
        assert v in [0, 1]
        return v


class CocoKeypointAnnotation(CocoInstanceAnnotation):
    keypoints: Keypoints
    num_keypoints: Optional[int] = None

    # make bbox optional
    bbox: Optional[Tuple[float, float, float, float]] = None

    @field_validator("keypoints")
    @classmethod
    def keypoints_must_be_multiple_of_three(cls, v: Keypoints) -> Keypoints:
        assert len(v) % 3 == 0, "keypoints list must be a multiple of 3"
        return v

    @field_validator("keypoints")
    @classmethod
    def keypoints_coordinates_must_be_in_pixel_space(cls, v: Keypoints) -> Keypoints:
        max_coordinate_value = 0.0
        for i in range(0, len(v), 3):
            max_coordinate_value = max(v[i], max_coordinate_value)
            max_coordinate_value = max(v[i + 2], max_coordinate_value)
        assert (
            max_coordinate_value > 1
        ), f"keypoints coordinates must be in pixel space, but max_coordinate is {max_coordinate_value}"
        return v

    @model_validator(mode="after")
    def num_keypoints_matches_amount_of_labeled_keypoints(self) -> "CocoKeypointAnnotation":
        labeled_keypoints = 0
        for v in self.keypoints[2::3]:
            if v > 0:
                labeled_keypoints += 1
        assert (
            labeled_keypoints == self.num_keypoints
        ), f"num_keypoints {self.num_keypoints} does not match number of labeled of keypoints {labeled_keypoints} for annotation {self.id}"
        return self


class CocoLicense(BaseModel):
    id: LicenseID
    name: str
    url: Url


class CocoInstancesDataset(BaseModel):
    info: Optional[CocoInfo] = None
    licenses: Optional[List[CocoLicense]] = None
    categories: Sequence[CocoCategory]
    images: List[CocoImage]
    annotations: Sequence[CocoInstanceAnnotation]

    @field_validator("annotations")
    @classmethod
    def annotations_list_cannot_be_empty(cls, v: List[CocoInstanceAnnotation]) -> List[CocoInstanceAnnotation]:
        assert len(v) > 0, "annotations list cannot be empty"
        return v

    # skip on failure becasue this validator requires the annotations list to be non-empty
    @model_validator(mode="after")
    def annotations_catergory_id_exist_in_categories(self) -> "CocoInstancesDataset":
        category_ids = set([category.id for category in self.categories])
        for annotation in self.annotations:
            assert (
                annotation.category_id in category_ids
            ), f"Annotation {annotation.id} has category_id {annotation.category_id} which does not exist in categories."
        return self


class CocoKeypointsDataset(CocoInstancesDataset):
    # Override the type of annotations.
    # annotations must be Sequence vs. list to allow this, see:
    # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
    categories: Sequence[CocoKeypointCategory]
    annotations: Sequence[CocoKeypointAnnotation]

    # skip on failure becasue this validator requires the annotations list to be non-empty
    @model_validator(mode="after")
    def num_keypoints_matches_annotations(self) -> "CocoKeypointsDataset":
        category_dict = {category.id: category for category in self.categories}
        for annotation in self.annotations:
            assert len(annotation.keypoints) // 3 == len(
                category_dict[annotation.category_id].keypoints
            ), f"Number of keypoints for annotation {annotation.id} does not match number of keypoints in category."
        return self



import json
import pathlib

import numpy as np
from airo_dataset_tools.data_parsers.pose import EulerAngles, Pose, Position


def test_pose_save_and_load(tmp_path: pathlib.Path):
    pose = Pose(
        position_in_meters=Position(x=1.0, y=2.0, z=3.0),
        rotation_euler_xyz_in_radians=EulerAngles(roll=np.pi / 4, pitch=-np.pi / 2, yaw=np.pi),
    )

    with open(tmp_path / "pose.json", "w") as file:
        json.dump(pose.model_dump(exclude_none=True), file, indent=4)

    with open(tmp_path / "pose.json", "r") as file:
        pose2 = Pose.model_validate_json(file.read())

    assert pose2.position_in_meters.x == 1.0
    assert pose2.position_in_meters.y == 2.0
    assert pose2.position_in_meters.z == 3.0
    assert pose2.rotation_euler_xyz_in_radians.roll == np.pi / 4
    assert pose2.rotation_euler_xyz_in_radians.pitch == -np.pi / 2
    assert pose2.rotation_euler_xyz_in_radians.yaw == np.pi


def test_homogeneous_coversion():
    pose = Pose(
        position_in_meters=Position(x=1.0, y=2.0, z=3.0),
        rotation_euler_xyz_in_radians=EulerAngles(roll=np.pi / 4, pitch=-np.pi / 3, yaw=np.pi / 6),  # no gimbal lock
    )

    pose_matrix = pose.as_homogeneous_matrix()
    pose2 = Pose.from_homogeneous_matrix(pose_matrix)

    # isclose is used because exact comparison failed due to tiny floating point rounding errors
    assert np.isclose(pose2.position_in_meters.x, pose.position_in_meters.x)
    assert np.isclose(pose2.position_in_meters.y, pose.position_in_meters.y)
    assert np.isclose(pose2.position_in_meters.z, pose.position_in_meters.z)
    assert np.isclose(pose2.rotation_euler_xyz_in_radians.roll, pose.rotation_euler_xyz_in_radians.roll)
    assert np.isclose(pose2.rotation_euler_xyz_in_radians.pitch, pose.rotation_euler_xyz_in_radians.pitch)
    assert np.isclose(pose2.rotation_euler_xyz_in_radians.yaw, pose.rotation_euler_xyz_in_radians.yaw)


if __name__ == "__main__":
    test_pose_save_and_load(pathlib.Path("."))
    test_homogeneous_coversion()



import json
import os
from typing import Any, Callable, List, Optional

import albumentations as A
import cv2
import numpy as np
import tqdm
from airo_dataset_tools.data_parsers.coco import (
    CocoImage,
    CocoInstanceAnnotation,
    CocoInstancesDataset,
    CocoKeypointAnnotation,
    CocoKeypointsDataset,
)
from airo_dataset_tools.segmentation_mask_converter import BinarySegmentationMask
from PIL import Image
from pydantic import ValidationError as PydanticValidationError


def apply_transform_to_coco_dataset(  # type: ignore # noqa: C901
    transforms: List[A.DualTransform],
    coco_dataset: CocoInstancesDataset,
    image_path: str,
    target_image_path: str,
    image_name_filter: Optional[Callable[[str], bool]] = None,
) -> CocoInstancesDataset:
    """Apply a sequence of albumentations transforms to a coco dataset, transforming images and keypoints, bounding boxes and segmentation masks if they are present.
    Present means that all annotations in the coco dataset have a bbox annotation.

    Args:
        transforms: _description_
        coco_dataset: _description_
        image_path: folder relative to which the image paths in the coco dataset are specified
        target_image_path: folder relative to which the image paths in the transformed coco dataset will be specified
        image_name_filter: optional filter for which images to transform based on their full path. Defaults to None.

    """
    # check if this is a keypoints dataset
    try:
        coco_dataset = CocoKeypointsDataset(**coco_dataset.model_dump(exclude_none=False))
        transform_keypoints = all(annotation.keypoints is not None for annotation in coco_dataset.annotations)

    except PydanticValidationError:
        transform_keypoints = False
    # check if bboxes and masks are present
    transform_bbox = all(annotation.bbox is not None for annotation in coco_dataset.annotations)
    transform_segmentation = all(annotation.segmentation is not None for annotation in coco_dataset.annotations)
    print(f"Transforming keypoints = {transform_keypoints}")
    print(f"Transforming bbox = {transform_bbox}")
    print(f"Transforming segmentation = {transform_segmentation}")

    keypoint_parameters = A.KeypointParams(format="xy", remove_invisible=False) if transform_keypoints else None
    bbox_parameters = A.BboxParams(format="coco", label_fields=["bbox_dummy_labels"]) if transform_bbox else None

    transform = A.Compose(
        transforms,
        keypoint_params=keypoint_parameters,
        bbox_params=bbox_parameters,
    )

    # create mappings between images & all corresponding annotations
    image_object_id_to_image_mapping = {image.id: image for image in coco_dataset.images}
    image_to_annotations_mapping: dict[CocoImage, List[CocoInstanceAnnotation]] = {
        coco_dataset.images[i]: [] for i in range(len(coco_dataset.images))
    }
    for annotation in coco_dataset.annotations:
        image_to_annotations_mapping[image_object_id_to_image_mapping[annotation.image_id]].append(annotation)

    for coco_image, annotations in tqdm.tqdm(image_to_annotations_mapping.items()):
        if image_name_filter is not None and image_name_filter(coco_image.file_name):
            print(f"skipping image {coco_image.file_name}")
            continue

        # load image
        pil_image = Image.open(os.path.join(image_path, coco_image.file_name)).convert(
            "RGB"
        )  # convert to RGB to avoid problems with PNG images
        image = np.array(pil_image)

        # combine annotations for all Annotation Instances related to the image
        # to transform them together with the image
        all_keypoints_xy = []
        all_bboxes = []
        all_masks = []
        for annotation in annotations:
            if transform_keypoints:
                assert isinstance(annotation, CocoKeypointAnnotation)
                all_keypoints_xy.extend(annotation.keypoints)
                # convert coco keypoints to list of (x,y) keypoints

            if transform_bbox:
                bbox = annotation.bbox
                assert bbox is not None  # for mypy
                # set bbox width, height to at least 1
                if bbox[3] < 1 or bbox[2] < 1:
                    # x_min must be < x_max for albumentations check
                    bbox = (0.0, 0.0, 1.0, 1.0)
                    print(
                        f"Invalid bbox for image {coco_image.file_name} and annotation {annotation.id}. Setting to [0.0, 0.0, 1.0, 1.0]"
                    )
                all_bboxes.append(bbox)

            if transform_segmentation:
                # convert segmentation to binary mask
                mask = annotation.segmentation
                assert mask is not None
                bitmap = BinarySegmentationMask.from_coco_segmentation_mask(
                    mask, coco_image.width, coco_image.height
                ).bitmap
                all_masks.append(bitmap)

        if transform_keypoints:
            all_keypoints_xy_nested = [all_keypoints_xy[i : i + 2] for i in range(0, len(all_keypoints_xy), 3)]

        arg_dict: dict[str, Any] = {
            "image": image,
        }
        if transform_keypoints:
            arg_dict["keypoints"] = all_keypoints_xy_nested
        if transform_bbox:
            arg_dict["bboxes"] = all_bboxes
            arg_dict["bbox_dummy_labels"] = [0 for _ in all_bboxes]
        if transform_segmentation:
            arg_dict["masks"] = all_masks
        transformed = transform(**arg_dict)

        # save transformed image
        transformed_image = transformed["image"]
        transformed_image = Image.fromarray(transformed_image)
        transformed_image_dir = os.path.join(target_image_path, os.path.dirname(coco_image.file_name))
        if not os.path.exists(transformed_image_dir):
            os.makedirs(transformed_image_dir)
        # specify quality to use for JPEG, (format is determined by file extension)
        # transformed_image.save(os.path.join(target_image_path, coco_image.file_name))
        # convert to BGR for opencv
        transformed_image_cv2 = cv2.cvtColor(np.array(transformed_image), cv2.COLOR_RGB2BGR)
        cv2.imwrite(os.path.join(target_image_path, coco_image.file_name), transformed_image_cv2)

        # change the metadata of the image coco object
        coco_image.width = transformed_image.width
        coco_image.height = transformed_image.height

        # store all modified annotations back into the coco dataset
        if transform_keypoints:
            all_transformed_keypoints_xy = transformed["keypoints"]
        if transform_bbox:
            all_transformed_bboxes = transformed["bboxes"]
        if transform_segmentation:
            all_transformed_masks = transformed["masks"]
        for annotation in annotations:
            if transform_keypoints:
                assert isinstance(annotation, CocoKeypointAnnotation)
                transformed_keypoints = all_transformed_keypoints_xy[: len(annotation.keypoints) // 3]
                all_transformed_keypoints_xy = all_transformed_keypoints_xy[len(annotation.keypoints) // 3 :]

                # set keypoints that are no longer in image to (0,0,0)
                for i, kp in enumerate(transformed_keypoints):
                    if 0 <= kp[0] < coco_image.width and 0 <= kp[1] < coco_image.height:
                        # add original visibility flag
                        transformed_keypoints[i] = [kp[0], kp[1], annotation.keypoints[i * 3 + 2]]
                    else:
                        transformed_keypoints[i] = [0.0, 0.0, 0]

                flattened_transformed_keypoints = [i for kp in transformed_keypoints for i in kp]
                annotation.keypoints = flattened_transformed_keypoints

            if transform_bbox:
                transformed_bbox = all_transformed_bboxes.pop(0)  # exactly one bbox per annotation
                annotation.bbox = transformed_bbox

            if transform_segmentation:
                transformed_segmentations = all_transformed_masks.pop(0)  # exactly one segmentation per annotation
                annotation.segmentation = BinarySegmentationMask(transformed_segmentations).as_polygon

    return coco_dataset


def resize_coco_dataset(
    annotations_json_path: str, width: int, height: int, target_dataset_dir: Optional[str] = None
) -> None:
    """Resize a COCO dataset. Will create a new directory at the `target_dataset_dir`with the resized dataset.
    Dataset is assumed to be
    /dir
        annotations.json # contains relative paths w.r.t. /dir
        ...
    """
    coco_dataset_dir = os.path.dirname(annotations_json_path)
    annotations_file_name = os.path.basename(annotations_json_path)
    dataset_parent_dir = os.path.dirname(coco_dataset_dir)
    if not target_dataset_dir:
        transformed_dataset_dir = os.path.join(
            dataset_parent_dir, f"{annotations_file_name.split('.')[0]}_resized_{width}x{height}"
        )
    else:
        transformed_dataset_dir = target_dataset_dir
    os.makedirs(transformed_dataset_dir, exist_ok=True)

    transforms = [A.Resize(height, width)]
    coco_json = json.load(open(annotations_json_path, "r"))
    coco_dataset = CocoInstancesDataset(**coco_json)
    transformed_dataset = apply_transform_to_coco_dataset(
        transforms, coco_dataset, coco_dataset_dir, transformed_dataset_dir
    )

    with open(os.path.join(transformed_dataset_dir, annotations_file_name), "w") as f:
        json.dump(transformed_dataset.model_dump(exclude_none=True), f)


if __name__ == "__main__":
    """example usage of the above function to resize all images in a coco dataset.
    Copy the following lines into your own codebase and modify as needed."""
    import pathlib

    path = pathlib.Path(__file__).parents[1] / "cvat_labeling" / "example" / "coco.json"

    coco_json_path = str(path)
    resize_coco_dataset(coco_json_path, 640, 480)



import pathlib

from airo_dataset_tools.cvat_labeling.load_xml_to_dict import get_dict_from_xml
from airo_dataset_tools.data_parsers.cvat_images import CVATImagesParser

path = pathlib.Path(__file__).parent.absolute()
CVAT_EXAMPLE_PATH = str(path.parent / "airo_dataset_tools" / "cvat_labeling" / "example" / "annotations.xml")
COCO_CATEGORIES_PATH = str(path.parent / "airo_dataset_tools" / "cvat_labeling" / "example" / "coco_categories.json")


def test_example_cvat_annotations_loading():
    cvat_dict = get_dict_from_xml(CVAT_EXAMPLE_PATH)
    cvat_keypoints_parser = CVATImagesParser(**cvat_dict)

    assert cvat_keypoints_parser.annotations.version == "1.1"
    assert len(cvat_keypoints_parser.annotations.image) == 4

    assert len(cvat_keypoints_parser.annotations.image[0].points) == 4  # 4 keypoints in first image



import json
import os

import pytest
from airo_dataset_tools.coco_tools.split_dataset import split_coco_dataset
from airo_dataset_tools.data_parsers.coco import CocoInstancesDataset


def test_keypoints_split():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_keypoints = CocoInstancesDataset(**data)
        datasets = split_coco_dataset(coco_keypoints, [0.5, 0.5], shuffle_before_splitting=False)
        assert len(datasets) == 2
        assert len(datasets[0].annotations) == 2
        assert len(datasets[1].annotations) == 3
        assert len(datasets[0].images) == 1
        assert len(datasets[1].images) == 1
        assert len(datasets[0].annotations[0].keypoints) > 0


def test_empty_annotations_split_raises_error():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    annotations = os.path.join(test_dir, "test_data/person_keypoints_val2017_small.json")

    with open(annotations, "r") as file:
        data = json.load(file)
        coco_keypoints = CocoInstancesDataset(**data)
        with pytest.raises(ValueError):
            split_coco_dataset(coco_keypoints, [0.9, 0.1], shuffle_before_splitting=False)



"""CLI interface for this package"""

import json
import os
from typing import List, Optional

import click
from airo_dataset_tools.coco_tools.change_coco_images_prefix import change_coco_json_image_prefix
from airo_dataset_tools.coco_tools.coco_instances_to_yolo import create_yolo_dataset_from_coco_instances_dataset
from airo_dataset_tools.coco_tools.fiftyone_viewer import view_coco_dataset
from airo_dataset_tools.coco_tools.merge_datasets import merge_coco_datasets
from airo_dataset_tools.coco_tools.split_dataset import split_and_save_coco_dataset
from airo_dataset_tools.coco_tools.transform_dataset import resize_coco_dataset
from airo_dataset_tools.cvat_labeling.convert_cvat_to_coco import cvat_image_to_coco


@click.group()
def cli() -> None:
    """CLI entrypoint for airo-dataset-tools"""


@cli.command(name="fiftyone-coco-viewer")  # no help, takes the docstring of the function.
@click.argument("annotations-json-path", type=click.Path(exists=True))
@click.option(
    "--dataset-dir",
    required=False,
    type=click.Path(exists=True),
    help="optional directory relative to which the image paths in the coco dataset are specified",
)
@click.option(
    "--label-types",
    "-l",
    multiple=True,
    type=click.Choice(["detections", "segmentations", "keypoints"]),
    help="add an argument for each label type you want to load (default: all)",
)
def view_coco_dataset_cli(
    annotations_json_path: str, dataset_dir: str, label_types: Optional[List[str]] = None
) -> None:
    """Explore COCO dataset with FiftyOne"""
    view_coco_dataset(annotations_json_path, dataset_dir, label_types)


@cli.command(name="convert-cvat-to-coco-keypoints")
@click.argument("cvat_xml_file", type=str, required=True)
@click.argument("coco-categories-json-file", type=str, required=True)
@click.option("--add-bbox", is_flag=True, default=False, help="include bounding box in coco annotations")
@click.option("--add-segmentation", is_flag=True, default=False, help="include segmentation in coco annotations")
def convert_cvat_to_coco_cli(
    cvat_xml_file: str, coco_categories_json_file: str, add_bbox: bool, add_segmentation: bool
) -> None:
    """Convert CVAT XML to COCO keypoints json according to specified coco categories"""
    coco = cvat_image_to_coco(
        cvat_xml_file, coco_categories_json_file, add_bbox=add_bbox, add_segmentation=add_segmentation
    )
    path = os.path.dirname(cvat_xml_file)
    filename = os.path.basename(cvat_xml_file)
    path = os.path.join(path, filename.split(".")[0] + ".json")
    with open(path, "w") as file:
        json.dump(coco, file)


@cli.command(name="resize-coco-dataset")
@click.argument("annotations-json-path", type=click.Path(exists=True))
@click.option("--width", type=int, required=True)
@click.option("--height", type=int, required=True)
@click.option("--target-dataset-dir", type=str, required=False)
def resize_coco_dataset_cli(
    annotations_json_path: str, width: int, height: int, target_dataset_dir: Optional[str]
) -> None:
    """Resize a COCO dataset. Will create a new directory with the resized dataset at the specified target_dataset_dir.
    Dataset is assumed to be
    /dir
        annotations.json # contains relative paths w.r.t. /dir
        ...
    """
    resize_coco_dataset(annotations_json_path, width, height, target_dataset_dir=target_dataset_dir)


@cli.command(name="coco-instances-to-yolo")
@click.option("--coco-json", type=str)
@click.option("--target-dir", type=str)
@click.option("--use-segmentation", is_flag=True)
def coco_intances_to_yolo(coco_json: str, target_dir: str, use_segmentation: bool) -> None:
    """Create a YOLO detections/segmentations dataset from a coco instances dataset"""
    print(f"converting coco instances dataset {coco_json} to yolo dataset {target_dir}")
    create_yolo_dataset_from_coco_instances_dataset(coco_json, target_dir, use_segmentation=use_segmentation)


@cli.command(name="split-coco-dataset")
@click.argument("json-path", type=click.Path(exists=True))
@click.option("--split-ratio", type=float, multiple=True, required=True)
@click.option("--shuffle-before-splitting", is_flag=True, default=True)
def split_coco_dataset_cli(json_path: str, split_ratio: List[float], shuffle_before_splitting: bool) -> None:
    """Split a COCO dataset into subsets according to the specified relative ratios and save them to disk.
    Images are split with their corresponding annotations. No guarantees on class balance or annotation ratios.

    If two ratios are specified, the dataset will be split into two subsets. these will be called train/val by default.
    If three ratios are specified, the dataset will be split into three subsets. these will be called train/val/test by default.

    e.g. split-coco-dataset <path> --split-ratio 0.8 --split-ratio 0.2
    """
    split_and_save_coco_dataset(json_path, split_ratio, shuffle_before_splitting=shuffle_before_splitting)


@cli.command(name="change-coco-images-prefix")
@click.argument("coco-json", type=click.Path(exists=True))
@click.option("--current-prefix", type=str, required=True)
@click.option("--new-prefix", type=str, required=True)
@click.option("--target-json-path", type=str)
def change_coco_images_prefix_cli(
    coco_json: str, current_prefix: str, new_prefix: str, target_json_path: Optional[str] = None
) -> None:
    """change the prefix of images in a coco dataset."""
    return change_coco_json_image_prefix(coco_json, current_prefix, new_prefix, target_json_path)


@cli.command(name="merge-coco-datasets")
@click.argument("coco-json-1", type=click.Path(exists=True))
@click.argument("coco-json-2", type=click.Path(exists=True))
@click.option(
    "--target-json-path",
    type=str,
    help="optional path to save the merged dataset to. If none is provided, a new directory will be created in the parent directory of coco-json-1 ",
)
def merge_coco_datasets_cli(coco_json_1: str, coco_json_2: str, target_json_path: Optional[str] = None) -> None:
    """merge two coco datasets into a single dataset."""
    if not target_json_path:
        target_json_path = os.path.join(os.path.dirname(coco_json_1), "merged")
        os.makedirs(target_json_path, exist_ok=True)
        target_json_path = os.path.join(target_json_path, "annotations.json")
    return merge_coco_datasets(coco_json_1, coco_json_2, target_json_path)


if __name__ == "__main__":
    cli()



""" module toconvert CVAT 1.1 images keypoint annotations to COCO Keypoint dataset"""

from __future__ import annotations

import json
from collections import defaultdict
from typing import List, Tuple

import tqdm
from airo_dataset_tools.cvat_labeling.load_xml_to_dict import get_dict_from_xml
from airo_dataset_tools.data_parsers.coco import (
    CocoImage,
    CocoKeypointAnnotation,
    CocoKeypointCategory,
    CocoKeypointsDataset,
)
from airo_dataset_tools.data_parsers.cvat_images import CVATImagesParser, ImageItem, LabelItem, Point
from airo_dataset_tools.segmentation_mask_converter import BinarySegmentationMask


def cvat_image_to_coco(  # noqa: C901, too complex
    cvat_xml_path: str, coco_configuration_json_path: str, add_bbox: bool = True, add_segmentation: bool = True
) -> dict:
    """Function that converts an annotation XML in the CVAT 1.1 Image format to the COCO keypoints format.
    If you don't need keypoints, you can simply use CVAT to create a COCOinstances format and should not use this function!


    The xml should be obtained with the AIRO-cvat-keypoints flow:
    It requires the CVAT dataset to be created by using labels formatted as <category>.<semantic_type>, using the group_id to group multiple instances together.
    if only a single instance is present, the group id is set to 1 by default so you don't have to do this yourself.

    Args:
        cvat_xml_path (str): _description_
        coco_configuration_json_path (str): path to the COCO categories to use for annotating this dataset.
        add_bbox (bool): add bounding box annotations to the COCO dataset, requires all keypoint annotations to have a bbox annotation
        add_segmentation (bool): add segmentation annotations to the COCO dataset, requires all keypoint annotations to have a mask annotation. Bboxes will be created from the segmentation masks.

    Returns: COCO Keypoints dataset model as a dict
    """
    cvat_dict = get_dict_from_xml(cvat_xml_path)
    cvat_parsed = CVATImagesParser(**cvat_dict)

    # create a COCO Dataset Model
    coco_images: List[CocoImage] = []
    coco_annotations: List[CocoKeypointAnnotation] = []
    coco_categories: List[CocoKeypointCategory] = []

    # load the COCO categories from the configuration file
    with open(coco_configuration_json_path, "r") as file:
        coco_categories_config = json.load(file)
    for category_dict in coco_categories_config["categories"]:
        category = CocoKeypointCategory(**category_dict)
        coco_categories.append(category)

    _validate_coco_categories_are_in_cvat(
        cvat_parsed, coco_categories, add_bbox=add_bbox, add_segmentation=add_segmentation
    )

    annotation_id_counter = 1  # counter for the annotation ID

    # iterate over all cvat annotations (grouped per image)
    # and create the COCO Keypoint annotations
    for cvat_image in tqdm.tqdm(cvat_parsed.annotations.image):
        coco_image = CocoImage(
            file_name=cvat_image.name,
            height=int(cvat_image.height),
            width=int(cvat_image.width),
            id=int(cvat_image.id) + 1,
        )
        coco_images.append(coco_image)
        for category in coco_categories:
            n_image_category_instances = _get_n_category_instances_in_image(cvat_image, category.name)
            for instance_id in range(1, n_image_category_instances + 1):  # IDs start with 1
                instance_category_keypoints = []
                for semantic_type in category.keypoints:
                    keypoint = _get_semantic_type_keypoint_for_instance_from_cvat_image(
                        cvat_image, semantic_type, instance_id
                    )
                    instance_category_keypoints.extend(keypoint)

                coco_annotations.append(
                    CocoKeypointAnnotation(
                        category_id=category.id,
                        id=annotation_id_counter,
                        image_id=coco_image.id,
                        keypoints=instance_category_keypoints,
                        num_keypoints=sum([1 if flag > 0 else 0 for flag in instance_category_keypoints[2::3]]),
                    )
                )
                if add_bbox:
                    coco_annotations[-1].bbox = _get_bbox_for_instance_from_cvat_image(cvat_image, instance_id)

                if add_segmentation:
                    coco_annotations[-1].segmentation = _get_segmentation_for_instance_from_cvat_image(
                        cvat_image, instance_id
                    )
                    coco_annotations[-1].iscrowd = 0
                    mask = BinarySegmentationMask.from_coco_segmentation_mask(
                        coco_annotations[-1].segmentation, coco_image.width, coco_image.height
                    )
                    coco_annotations[-1].area = mask.area

                    if not add_bbox:
                        coco_annotations[-1].bbox = mask.bbox

                annotation_id_counter += 1

    coco_model = CocoKeypointsDataset(images=coco_images, annotations=coco_annotations, categories=coco_categories)
    return coco_model.model_dump(exclude_none=True)


####################
### helper functions
####################


def _validate_coco_categories_are_in_cvat(
    cvat_parsed: CVATImagesParser, coco_categories: List[CocoKeypointCategory], add_bbox: bool, add_segmentation: bool
) -> None:
    # gather the annotation from CVAT
    cvat_categories_dict = defaultdict(list)

    labels = cvat_parsed.annotations.meta.get_job_or_task().labels.label

    # Handle both single LabelItem and list of LabelItems
    if isinstance(labels, LabelItem):
        labels = [labels]

    for annotation_category in labels:
        assert isinstance(annotation_category, LabelItem)
        category_str, annotation_name = annotation_category.name.split(".")
        cvat_categories_dict[category_str].append(annotation_name)

    for category_str, semantic_types in cvat_categories_dict.items():
        if add_bbox:
            assert "bbox" in semantic_types, "bbox annotations are required"
        if add_segmentation:
            assert "mask" in semantic_types, "segmentation masks are required"
        # find the matching COCO category
        coco_category = None
        for coco_category in coco_categories:
            if coco_category.name == category_str:
                break
        if coco_category is not None:
            for category_keypoint in coco_category.keypoints:
                assert category_keypoint in semantic_types, f"semantic type {category_keypoint} not found"
        else:
            raise AssertionError(
                f"category {category_str} not found in coco categories. "
                f"Available categories: {[c.name for c in coco_categories]}"
            )


def _get_n_category_instances_in_image(cvat_image: ImageItem, category_name: str) -> int:
    """returns the number of instances for the specified category in the CVAT ImageItem.

    This is done by finding the maximum group_id for all annotations of the image.

    Edge cases include: no Points in the image or only 1 Point in the image.
    """
    if cvat_image.points is None:
        return 0
    if not isinstance(cvat_image.points, list):
        if _get_category_from_cvat_label(cvat_image.points.label) == category_name:
            assert cvat_image.points.group_id is not None, "group_id was None"
            return int(cvat_image.points.group_id)
        else:
            return 0
    max_group_id = 0
    for cvat_point in cvat_image.points:
        if _get_category_from_cvat_label(cvat_point.label) == category_name:
            assert cvat_point.group_id is not None, "group_id was None"
            max_group_id = max(max_group_id, int(cvat_point.group_id))
    return max_group_id


def _get_category_from_cvat_label(label: str) -> str:
    """cvat labels are formatted as <category>.<semantic_type>
    this function returns the category
    """
    split = label.split(".")
    assert len(split) == 2, " label was not formatted as category.semantic_type"
    return label.split(".")[0]


def _get_semantic_type_from_cvat_label(label: str) -> str:
    """cvat labels are formatted as <category>.<semantic_type>
    this function returns the semantic type
    """
    split = label.split(".")
    assert len(split) == 2, " label was not formatted as category.semantic_type"
    return label.split(".")[1]


def _get_bbox_for_instance_from_cvat_image(
    cvat_image: ImageItem, instance_id: int
) -> Tuple[float, float, float, float]:
    """returns the bbox for the instance in the cvat image."""
    instance_id_str = str(instance_id)
    if cvat_image.box is None:
        raise ValueError("bbox annotations are required for image {cvat_image.name}")
    if not isinstance(cvat_image.box, list):
        if instance_id_str == cvat_image.box.group_id:
            return (
                float(cvat_image.box.xtl),
                float(cvat_image.box.ytl),
                float(cvat_image.box.xbr) - float(cvat_image.box.xtl),
                float(cvat_image.box.ybr) - float(cvat_image.box.ytl),
            )
        else:
            raise ValueError("bbox annotations are required for image {cvat_image.name}")
    for bbox in cvat_image.box:
        if instance_id_str == bbox.group_id:
            return (
                float(bbox.xtl),
                float(bbox.ytl),
                float(bbox.xbr) - float(bbox.xtl),
                float(bbox.ybr) - float(bbox.ytl),
            )
    raise ValueError("bbox annotations are required for image {cvat_image.name}")


def _get_segmentation_for_instance_from_cvat_image(cvat_image: ImageItem, instance_id: int) -> List[List[float]]:
    """returns the segmentation polygon for the instance in the cvat image."""
    instance_id_str = str(instance_id)
    if cvat_image.polygon is None:
        raise ValueError(f"segmentation annotations are required for image {cvat_image.name}")
    if not isinstance(cvat_image.polygon, list):
        if instance_id_str == cvat_image.polygon.group_id:
            polygon_str = cvat_image.polygon.points
            polygon_str = polygon_str.replace(";", ",")
            return [[float(x) for x in polygon_str.split(",")]]
        else:
            raise ValueError("segmentation annotations are required for image {cvat_image.name}")
    for polygon in cvat_image.polygon:
        if instance_id_str == polygon.group_id:
            polygon_str = polygon.points
            polygon_str = polygon_str.replace(";", ",")
            return [[float(x) for x in polygon_str.split(",")]]
    raise ValueError("segmentation annotations are required for image {cvat_image.name}")


def _get_semantic_type_keypoint_for_instance_from_cvat_image(
    cvat_image: ImageItem, semantic_type: str, instance_id: int
) -> List[float]:
    """Finds the keypoint of the given semantic type for this instance in the image.
    Returns [0,0,0] if the keypoint is not annotated for this instance.

    Args:
        cvat_image (ImageItem): _description_
        semantic_type (str): _description_
        instance_id (int): _description_

    Returns:
        List: [x,y,visibility]
    """
    instance_id_str = str(instance_id)
    if cvat_image.points is None:
        return [0.0, 0.0, 0]
    if not isinstance(cvat_image.points, list):
        if (
            semantic_type == _get_semantic_type_from_cvat_label(cvat_image.points.label)
            and instance_id_str == cvat_image.points.group_id
        ):
            return _extract_coco_keypoint_from_cvat_point(cvat_image.points)
        else:
            return [0.0, 0.0, 0]
    for cvat_point in cvat_image.points:
        if (
            semantic_type == _get_semantic_type_from_cvat_label(cvat_point.label)
            and instance_id_str == cvat_point.group_id
        ):
            return _extract_coco_keypoint_from_cvat_point(cvat_point)
    return [0.0, 0.0, 0]


def _extract_coco_keypoint_from_cvat_point(cvat_point: Point) -> List:
    """extract keypoint in coco format (u,v,f) from cvat annotation point.
    Args:
        cvat_point (Point): _description_

    Returns:
        List: [u,v,f] where u,v are the coords scaled to the image resolution and f is the coco visibility flag.
        see the coco dataset format for more details.
    """
    u = float(cvat_point.points.split(",")[0])
    v = float(cvat_point.points.split(",")[1])
    f = (
        1 if cvat_point.occluded == "1" else 2
    )  # occluded = 1 means not visible, which is 1 in COCO; visible in COCO is 2
    return [u, v, f]


if __name__ == "__main__":
    """
    For development. See cli.py for the command line interface to this function that you can use to convert cvat annotations to coco.
    """

    import pathlib

    path = pathlib.Path(__file__).parent.absolute()
    cvat_xml_file = str(path / "example" / "annotations.xml")

    coco_categories_file = str(path / "example" / "coco_categories.json")

    coco = cvat_image_to_coco(cvat_xml_file, coco_categories_file, add_bbox=True, add_segmentation=False)
    with open("coco.json", "w") as file:
        json.dump(coco, file)



import json
import os
import pathlib

from airo_dataset_tools.coco_tools.merge_datasets import merge_coco_datasets
from airo_dataset_tools.coco_tools.split_dataset import split_and_save_coco_dataset
from airo_dataset_tools.data_parsers.coco import CocoKeypointsDataset


def test_split_and_merge():
    dataset_path = pathlib.Path(__file__).parent / "test_data" / "lego-battery-resized" / "annotations.json"
    split_and_save_coco_dataset(str(dataset_path), [0.6, 0.4])
    json1 = pathlib.Path(__file__).parent / "test_data" / "lego-battery-resized" / "annotations_train.json"
    json2 = pathlib.Path(__file__).parent / "test_data" / "lego-battery-resized" / "annotations_val.json"
    target_json = (
        pathlib.Path(__file__).parent / "test_data" / "lego-battery-resize-merged" / "annotations_merged.json"
    )
    merged_dataset = merge_coco_datasets(str(json1), str(json2), str(target_json))
    assert target_json.exists()
    merged_dataset = CocoKeypointsDataset(**json.load(target_json.open("r")))
    original_dataset = CocoKeypointsDataset(**json.load(dataset_path.open("r")))
    # some basic checks to see if the merge was successful
    assert len(merged_dataset.images) == len(original_dataset.images)
    assert len(merged_dataset.annotations) == len(original_dataset.annotations)
    assert len(merged_dataset.categories) == len(original_dataset.categories)

    os.remove(json1)
    os.remove(json2)
    os.remove(target_json)



from __future__ import annotations

import numpy as np
from airo_spatial_algebra.se3 import SE3Container
from airo_typing import HomogeneousMatrixType
from pydantic import BaseModel


class Position(BaseModel):
    """Position in 3D space, all units are in meters."""

    x: float
    y: float
    z: float


class EulerAngles(BaseModel):
    roll: float
    pitch: float
    yaw: float


class Pose(BaseModel):
    """Pose of an object in 3D space, all units are in meters and radians.

    The euler angles are  extrinsic (rotations about the axes xyz of the original coordinate system, which is assumed
    to remain motionless).
    """

    position_in_meters: Position
    rotation_euler_xyz_in_radians: EulerAngles

    @classmethod
    def from_homogeneous_matrix(cls, matrix: HomogeneousMatrixType) -> Pose:
        """Creates a Pose object from a 4x4 homogeneous transformation matrix."""
        se3_pose = SE3Container.from_homogeneous_matrix(matrix)
        position = se3_pose.translation
        euler_angles = se3_pose.orientation_as_euler_angles

        position_model = Position(x=position[0], y=position[1], z=position[2])
        euler_angles_model = EulerAngles(roll=euler_angles[0], pitch=euler_angles[1], yaw=euler_angles[2])

        pose = cls(position_in_meters=position_model, rotation_euler_xyz_in_radians=euler_angles_model)
        return pose

    def as_homogeneous_matrix(self) -> HomogeneousMatrixType:
        """Returns the pose as a 4x4 homogeneous transformation matrix."""
        position = self.position_in_meters
        euler_angles = self.rotation_euler_xyz_in_radians

        position_array = np.array([position.x, position.y, position.z])
        euler_angles_array = np.array([euler_angles.roll, euler_angles.pitch, euler_angles.yaw])

        pose_matrix = SE3Container.from_euler_angles_and_translation(
            euler_angles_array, position_array
        ).homogeneous_matrix

        return pose_matrix



import json
import pathlib

from airo_dataset_tools.data_parsers.camera_intrinsics import (
    CameraIntrinsics,
    FocalLengths,
    PrincipalPoint,
    Resolution,
)


def test_camera_intrinsics_save_and_load(tmp_path: pathlib.Path):
    """
    Based on the follwing intrinsics of one of our ZED2i cameras:
    [LEFT_CAM_2K]
    fx=1067.91
    fy=1068.05
    cx=1107.48
    cy=629.675
    k1=-0.0542749
    k2=0.0268096
    p1=0.000204483
    p2=-0.000310015
    k3=-0.0104089
    """
    camera_intrinsics = CameraIntrinsics(
        image_resolution=Resolution(width=2208, height=1520),
        focal_lengths_in_pixels=FocalLengths(fx=1067.91, fy=1068.05),
        principal_point_in_pixels=PrincipalPoint(cx=1107.48, cy=629.675),
        radial_distortion_coefficients=[-0.0542749, 0.0268096, -0.0104089],
        tangential_distortion_coefficients=[0.000204483, -0.000310015],
    )

    with open(tmp_path / "camera_intrinsics.json", "w") as file:
        json.dump(camera_intrinsics.model_dump(exclude_none=True), file, indent=4)

    with open(tmp_path / "camera_intrinsics.json", "r") as file:
        camera_intrinsics2 = CameraIntrinsics.model_validate_json(file.read())

    assert camera_intrinsics2.image_resolution.width == 2208
    assert camera_intrinsics2.image_resolution.height == 1520
    assert camera_intrinsics2.focal_lengths_in_pixels.fx == 1067.91
    assert camera_intrinsics2.focal_lengths_in_pixels.fy == 1068.05
    assert camera_intrinsics2.principal_point_in_pixels.cx == 1107.48
    assert camera_intrinsics2.principal_point_in_pixels.cy == 629.675
    assert camera_intrinsics2.radial_distortion_coefficients[0] == -0.0542749
    assert camera_intrinsics2.radial_distortion_coefficients[1] == 0.0268096
    assert camera_intrinsics2.radial_distortion_coefficients[2] == -0.0104089
    assert camera_intrinsics2.tangential_distortion_coefficients[0] == 0.000204483
    assert camera_intrinsics2.tangential_distortion_coefficients[1] == -0.000310015


if __name__ == "__main__":
    test_camera_intrinsics_save_and_load(pathlib.Path("."))



import json
from typing import Optional

from airo_dataset_tools.data_parsers.coco import CocoInstancesDataset


def change_coco_dataset_image_prefix(
    coco_dataset: CocoInstancesDataset, current_prefix: str, target_prefix: str
) -> CocoInstancesDataset:
    """Change the prefix of the image paths in a COCO dataset. Can be used to change the base directory of the images in a dataset.

    Args:
        coco_dataset: the dataset to modify
        current_prefix: the current prefix of the image paths in the dataset
        target_prefix: the target prefix of the image paths in the dataset

    e.g.

    if the images are currently relative to the image folder inside the dataset and you want to make them relative to the dataset folder:
    change_coco_image_base_dir(coco_dataset, "", "images/")
    """

    for image in coco_dataset.images:
        image.file_name = image.file_name.removeprefix(current_prefix)
        image.file_name = target_prefix + image.file_name

    return coco_dataset


def change_coco_json_image_prefix(
    coco_json_file: str, current_prefix: str, target_prefix: str, target_json_file: Optional[str]
) -> None:
    with open(coco_json_file, "r") as f:
        coco_dataset = CocoInstancesDataset(**json.load(f))

    coco_dataset = change_coco_dataset_image_prefix(coco_dataset, current_prefix, target_prefix)
    if target_json_file is None:
        target_json_file = coco_json_file
    with open(target_json_file, "w") as f:
        json.dump(coco_dataset.model_dump(), f)



import pathlib

import setuptools

root_folder = pathlib.Path(__file__).parents[1]
setuptools.setup(
    name="airo-dataset-tools",
    version="2025.8.0",
    author="Victor-Louis De Gusseme",
    author_email="victorlouisdg@gmail.com",
    description="Scripts for loading and converting datasets for the Ghent University AI and Robotics Lab",
    install_requires=[
        "numpy>=2.0",
        "pydantic>2.0.0",  # pydantic 2.0.0 has a lot of breaking changes
        "opencv-contrib-python==4.10.0.84",  # See airo-camera-toolkit setup.py for explanation
        "opencv-python-headless==4.10.0.84",  # See airo-camera-toolkit setup.py for explanation
        "pycocotools",
        "xmltodict",
        "tqdm",
        "fiftyone",  # visualization
        "Pillow",
        "types-Pillow",
        "albumentations",
        "click",
        "airo-typing>=2025.8.0",
        "airo-spatial-algebra>=2025.8.0",
    ],
    packages=setuptools.find_packages(exclude=["test"]),
    entry_points={
        "console_scripts": [
            "airo-dataset-tools = airo_dataset_tools.cli:cli",
        ],
    },
    package_data={"airo_dataset_tools": ["py.typed"]},
)



import numpy as np
from airo_dataset_tools.coco_tools.transforms import PillowResize


def test_resize():
    img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    resize = PillowResize(10, 10)
    resized_img = resize(image=img)["image"]
    assert resized_img.shape == (10, 10, 3)



from typing import Any

import albumentations as A
import numpy as np
from PIL import Image


class PillowResize(A.Resize):  # type: ignore
    """Use Pillow (instead of OpenCV) to resize the input to the given height and width.
    always uses Bicubic interpolation.

    PIllow instead of Opencv because opencv does not adapt the filter size to the scaling factor,
    which creates artifacts in the output for large downscaling factors.
    cf. https://arxiv.org/pdf/2104.11222.pdf

    Args:
        height (int): desired height of the output.
        width (int): desired width of the output.
        p (float): probability of applying the transform. Default: 1.

    Targets:
        image, mask, bboxes, keypoints

    Image types:
        uint8, float32
    """

    def __init__(
        self, height: int, width: int, interpolation: Any = Image.BICUBIC, always_apply: bool = False, p: float = 1.0
    ):
        super(PillowResize, self).__init__(height, width, always_apply=always_apply, p=p)
        self.height = height
        self.width = width
        self.interpolation = interpolation

    def apply(self, img: np.ndarray, **params: dict) -> np.ndarray:
        return np.array(Image.fromarray(img).resize((self.width, self.height), self.interpolation))



import json
import pathlib
from collections import defaultdict

import cv2
import tqdm
from airo_dataset_tools.data_parsers.coco import CocoInstancesDataset
from airo_dataset_tools.segmentation_mask_converter import BinarySegmentationMask


def create_yolo_dataset_from_coco_instances_dataset(
    coco_dataset_json_path: str, target_directory: str, use_segmentation: bool = False
) -> None:
    """Converts a coco dataset to a yolo dataset, either segmentations or detections


    coco dataset is expected:
    dataset/
        images/
        <>.json # containing paths relative to /dataset

    yolo dataset format will be

    dataset/
        images/
        labels/
        <>.txt # paths as in coco dataset /images subdir.

        # ultralytics uses an additional <>.json, this has to be created manually..
        # we do export a .names file, as in the darknet yolo format.
        # each line contains a class name, the line number is the class id
        obj.names
    Args:
        coco_dataset_json_path: _description_
        target_directory: _description_
        use_segmentation: create a segmentation dataset instead of a detection dataset (requires segmentation annotations)


    quiz question: how many people would have written a similar converter in the past?...

    """
    _coco_dataset_json_path = pathlib.Path(coco_dataset_json_path)
    _target_directory = pathlib.Path(target_directory)

    # load the json file into the coco dataset object
    annotations = json.load(open(_coco_dataset_json_path, "r"))
    coco_dataset = CocoInstancesDataset(**annotations)

    target_dir = pathlib.Path(_target_directory)
    target_dir.mkdir(parents=True, exist_ok=True)
    image_dir = target_dir / "images"
    label_dir = target_dir / "labels"
    image_dir.mkdir(parents=True, exist_ok=True)
    label_dir.mkdir(parents=True, exist_ok=True)

    # create the yolo dataset
    annotations = coco_dataset.annotations
    image_id_to_image = {image.id: image for image in coco_dataset.images}
    image_id_to_annotations = defaultdict(list)
    for annotation in annotations:
        image_id_to_annotations[annotation.image_id].append(annotation)

    category_id_to_category = {category.id: category for category in coco_dataset.categories}
    # sort by original COCO ID, but whereas COCO does not care about ID ranges, YOLO expects them to be in the range 0..N-1
    # so we sort them by ID and then use the index as the new ID
    yolo_category_index = list(sorted(category_id_to_category.values(), key=lambda category: category.id))

    for image_id, annotations in tqdm.tqdm(image_id_to_annotations.items()):
        coco_image = image_id_to_image[image_id]
        image_path = pathlib.Path(coco_image.file_name)
        if not image_path.is_absolute():
            image_path = _coco_dataset_json_path.parent / image_path

        relative_image_path = image_path.relative_to(_coco_dataset_json_path.parent)

        # ultralytics parser finds the 'latest' occurance of 'images' in the dataset path,
        #  so we need to replace any occurance of 'image' with 'img'
        #  https://github.com/ultralytics/ultralytics/issues/3581

        relative_image_path_str = str(relative_image_path).replace("image", "img")
        relative_image_path = pathlib.Path(relative_image_path_str)

        image = cv2.imread(str(image_path))
        height, width, _ = image.shape

        label_path = label_dir / f"{relative_image_path.with_suffix('')}.txt"
        label_path.parent.mkdir(parents=True, exist_ok=True)

        with open(label_path, "w") as file:
            for annotation in annotations:
                category = category_id_to_category[annotation.category_id]
                yolo_id = yolo_category_index.index(category)
                if use_segmentation:
                    segmentation = annotation.segmentation
                    # convert to **single** polygon
                    segmentation = BinarySegmentationMask.from_coco_segmentation_mask(segmentation, width, height)
                    segmentation = segmentation.as_single_polygon

                    if segmentation is None:
                        # should actually never happen as each annotation is assumed to have a segmentation if you pass use_segmentation=True
                        # but we filter it for convenience to deal with edge cases
                        print(f"skipping annotation for image {image_path}, as it has no segmentation")
                        continue

                    file.write(f"{yolo_id}")
                    for (x, y) in zip(segmentation[0::2], segmentation[1::2]):
                        file.write(f" {x/width} {y/height}")
                    file.write("\n")

                else:
                    x, y, w, h = annotation.bbox
                    x_center = x + w / 2
                    y_center = y + h / 2
                    x_center /= width
                    y_center /= height
                    w /= width
                    h /= height
                    file.write(f"{yolo_id} {x_center} {y_center} {w} {h}\n")

        image_target_path = image_dir / relative_image_path
        image_target_path.parent.mkdir(parents=True, exist_ok=True)
        cv2.imwrite(str(image_target_path), image)

    # create the obj.names file
    with open(target_dir / "obj.names", "w") as file:
        # sort categories by id
        for category in yolo_category_index:
            file.write(f"{category.name}\n")


if __name__ == "__main__":
    """example usage"""
    import os

    path = pathlib.Path(__file__).parents[1] / "cvat_labeling" / "example" / "coco.json"

    coco_json_path = str(path)
    coco_dir = os.path.dirname(coco_json_path)
    coco_file_name = os.path.basename(coco_json_path)
    coco_target_dir = os.path.join(os.path.dirname(coco_dir), "yolo_dataset")
    os.makedirs(coco_target_dir, exist_ok=True)

    create_yolo_dataset_from_coco_instances_dataset(coco_json_path, coco_target_dir, use_segmentation=True)



from __future__ import annotations

from typing import List, Optional, Tuple

import numpy as np
from airo_typing import CameraIntrinsicsMatrixType, CameraResolutionType
from pydantic import BaseModel


class Resolution(BaseModel):
    width: int
    height: int

    def as_tuple(self) -> CameraResolutionType:
        return self.width, self.height


class FocalLengths(BaseModel):
    fx: float
    fy: float


class PrincipalPoint(BaseModel):
    cx: float
    cy: float


RadialDistortionCoefficients = List[float]
TangentialDistortionCoefficients = List[float]


class CameraIntrinsics(BaseModel):
    """A format for storing the camera intrinsics at a specific image resolution."""

    image_resolution: Resolution
    focal_lengths_in_pixels: FocalLengths
    principal_point_in_pixels: PrincipalPoint

    # Distortion coefficients are stored so you can add as many as you want.
    radial_distortion_coefficients: Optional[RadialDistortionCoefficients] = None
    tangential_distortion_coefficients: Optional[TangentialDistortionCoefficients] = None

    @classmethod
    def from_matrix_and_resolution(
        cls, intrinsics_matrix: CameraIntrinsicsMatrixType, resolution: Tuple[int, int]
    ) -> CameraIntrinsics:
        """Creates a CameraIntrinsics object from a 3x3 matrix and an image resolution (width, height)."""
        fx = intrinsics_matrix[0, 0]
        fy = intrinsics_matrix[1, 1]
        cx = intrinsics_matrix[0, 2]
        cy = intrinsics_matrix[1, 2]

        width, height = resolution

        camera_intrinsics = cls(
            image_resolution=Resolution(width=width, height=height),
            focal_lengths_in_pixels=FocalLengths(fx=fx, fy=fy),
            principal_point_in_pixels=PrincipalPoint(cx=cx, cy=cy),
        )
        return camera_intrinsics

    def as_matrix(self) -> CameraIntrinsicsMatrixType:
        """Returns the camera intrinsics as a 3x3 matrix, often called K."""
        fx = self.focal_lengths_in_pixels.fx
        fy = self.focal_lengths_in_pixels.fy
        cx = self.principal_point_in_pixels.cx
        cy = self.principal_point_in_pixels.cy

        intrinsics_matrix = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])
        return intrinsics_matrix
